
#define __RGB_IMPORT 1
#include "rgbnodeutils.h"
#include "zlib.h"

#undef  __RGB_IMPORT
#define __RGB_EXPORT 1
#include "rgbos.h"  /* Redefine IMPORTEXPORT for exporting symbols */

/* The specialString parameter allows the user to provide a string in the
   control center for the transformation algorithms to use.

   The intent of this string is to provide a key for encryption/decryption
   etc. However, it could be used to influence compression technique
   (i.e. it would be trivial to use it to allow the user to specify
   preference for speed over compression ratio, etc.   
*/

void __RGB_IMPORTEXPORTPREFIX __RGB_IMPORTEXPORTSUFFIX compressData(struct CciByteArray* inputData,
                  struct CciByteArray* outputData, CciChar* specialString)
{
      static char* functionName = (char*)"compressData()";
      z_stream compressor;
      int rc=0;
      int iRc;
      compLog(1, functionName);

      compressor.zalloc = Z_NULL;
      compressor.zfree = Z_NULL;
      compressor.opaque = Z_NULL;
      
      compLog(0,"Initialising compressor");
      rc = deflateInit(&compressor, Z_DEFAULT_COMPRESSION);
      if (rc != Z_OK)
	{
	  cciThrowException(&rc,E_RECOV,RGB_COMPRESS_INIT_FAIL,
			    "Failed to initialise compressor/decompressor "
			    "object.", (char*)0);
	}
      
      compressor.next_in = (Bytef*)inputData->pointer;
      compressor.avail_in = (uInt)inputData->size;
      compressor.next_out = (Bytef*)outputData->pointer;
      compressor.avail_out = (uInt)(outputData->size);
      
      compLog(0,"Starting compression with avail_in %d and avail_out %d",
	      compressor.avail_in,
	      compressor.avail_out);

      while ((rc = deflate(&compressor, Z_FINISH)) 
	     != Z_STREAM_END)
	{
	  /*  either an error or need more output space - in which  */
	  /*  case the deflation has caused an inflation! */
	  
	  compLog(0,"deflate not yet complete. Return code: %d.",
		  rc);

	  if (rc == Z_OK)
	    {
	      /*  provide more output space */
	      compLog(0,"Data expanded under compression! Allocating "
		      "more space");
	      
	      if (compressor.avail_out != 0)
		{
		  /*  if this occurs, my understanding of the zlib library */
		  /*  is suspect... */
		  cciThrowException(&rc,E_RECOV,RGB_INCONSISTENT_COMP_STATE,
				    "The (de)compressor is in an invalid "
				    "state.",
				    "Expected avail_out to be zero.",
				    (char*)0);
		}
	      
	      outputData->pointer=realloc(outputData->pointer,
					 outputData->size+inputData->size+50);
	      TESTMALLOC(outputData->pointer);
	      
	      compLog(0,"Successfully reallocated %d bytes.",
		      outputData->size+inputData->size+50);

	      /*  point next_out location to be one after the whole size */
	      /*  of the previous memory chunk (that's why we set this */
	      /*  before increasing the outputData.size variable) */
	      
	      compressor.next_out=(unsigned char*)outputData->pointer+outputData->size;
	      compressor.avail_out=inputData->size+50;
	      
	      compLog(0,"avail_out is now %d.",
		      compressor.avail_out);
	      
	      /*  The adding of '50' is purely  */
	      /*  arbitrary. It is to prevent problems when inputDataSize */
	      /*  is zero - I don't want any silly infinite loops */
	      
	      outputData->size+=(inputData->size+50);
	    }
	  else
	    {
	      compLog(0, "compressor did not return Z_OK. Returned: %d", rc);
	      cciThrowException(&iRc,E_RECOV,RGB_ZLIB_FAILURE,
				"The compressor returned a non Z_OK "
				"code.", "None Z_OK return code", (char*)0);
	    }
	}
      
      compLog(0, "Deflate successfully ended");

      /*  update output buffer size field to represent actual
	  size of output data */
      outputData->size = outputData->size-(int)compressor.avail_out;
      deflateEnd(&compressor); /*  don't care about the return value */

      compLog(-1, functionName);
      return;
}

void __RGB_IMPORTEXPORTPREFIX __RGB_IMPORTEXPORTSUFFIX decompressData(struct CciByteArray* inputData,
		    struct CciByteArray* outputData, CciChar* specialString)
{
      static char* functionName = (char*)"compressData()";
      z_stream decompressor;
      int rc;
      int iRc;

      compLog(1, functionName);

      decompressor.zalloc = Z_NULL;
      decompressor.zfree = Z_NULL;
      decompressor.opaque = Z_NULL;

      decompressor.next_in = (Bytef*)inputData->pointer;
      decompressor.avail_in = (uInt)(inputData->size);
      decompressor.next_out = (Bytef*)outputData->pointer;
      decompressor.avail_out = (uInt)outputData->size;
      
      compLog(0, "Initialising decompressor");
      rc = inflateInit(&decompressor);
      if (rc != Z_OK)
	{
	  cciThrowException(&iRc,E_RECOV,RGB_COMPRESS_INIT_FAIL,
			    "Failed to initialise decompressor object",
			    "Failed to initialise decompressor object",
			    (char*)0);
	}
      
      compLog(0, "Inflating data...");

      while ((rc = inflate(&decompressor, Z_NO_FLUSH)) 
	     != Z_STREAM_END)
	{
	  /*  either an error or need more output space - in which  */
	  /*  case the deflation has caused an inflation! */
	  
	  compLog(0,"inflate not yet complete. Return code: %d.",
		  rc);

	  if (rc == Z_OK)
	    {
	      /*  provide more output space */
	      compLog(0,"Need more output space... Allocating ");
		      	      
	      if (decompressor.avail_out != 0)
		{
		  /*  if this occurs, my understanding of the zlib library */
		  /*  is suspect... */
		  cciThrowException(&rc,E_RECOV,RGB_INCONSISTENT_COMP_STATE,
				    "The (de)compressor is in an invalid "
				    "state.",
				    "Expected avail_out to be zero.",
				    (char*)0);
		}
	      
	      outputData->pointer=realloc(outputData->pointer,
					  outputData->size+inputData->size+50);
	      TESTMALLOC(outputData->pointer);
	      
	      compLog(0,"Successfully reallocated %d bytes.",
		      outputData->size+inputData->size+50);

	      /*  point next_out location to be one after the whole size */
	      /*  of the previous memory chunk (that's why we set this */
	      /*  before increasing the outputData.size variable) */
	      
	      decompressor.next_out=(unsigned char*)outputData->pointer+outputData->size;
	      decompressor.avail_out=inputData->size+50;
	      
	      compLog(0,"avail_out is now %d.",
		      decompressor.avail_out);
	      
	      /*  The adding of '50' is purely  */
	      /*  arbitrary. It is to prevent problems when inputDataSize */
	      /*  is zero - I don't want any silly infinite loops */
	      	      
	      outputData->size+=(inputData->size+50);
	    }
	  else
	    {
	      compLog(0, "Decompressor returned no Z_OK return code: %d", rc);
	      cciThrowException(&iRc,E_RECOV,RGB_ZLIB_FAILURE,
				"The compressor returned a non Z_OK "
				"code.", "None Z_OK return code when "
				"attempting to decompress data. Was the "
				"input message generated by the ZLIB library?", (char*)0);
	    }
	}


      compLog(0, "... successful");
      outputData->size = outputData->size-(int)decompressor.avail_out;
      inflateEnd(&decompressor); /*  don't care about the return value */

      compLog(-1, functionName);
      return;
}

